from dataclasses import dataclass, field, replace

from bussdcc_system.model import DeviceSpec

from .settings import Settings


@dataclass(slots=True, frozen=True)
class Config:
    settings: Settings = field(
        metadata={
            "label": "Settings",
            "group": "General",
            "required": True,
        }
    )
    devices: dict[str, DeviceSpec] = field(
        default_factory=dict,
        metadata={
            "label": "Devices",
            "group": "Runtime",
            "help": "Named device registrations.",
        },
    )

    def normalized(self) -> "Config":
        return replace(self, settings=self.settings.normalized())



from typing import Any
from flask import Blueprint, render_template, redirect, url_for, request

from bussdcc_framework.interface.web import current_ctx
from bussdcc_framework.codec import load_value

from .... import message, config

from bussdcc_framework.interface.web import formtree

bp = Blueprint("settings", __name__, url_prefix="/settings")


@bp.route("/")
def index() -> Any:
    ctx = current_ctx()
    settings = ctx.state.get("settings")
    if settings is None:
        return redirect(url_for("settings.new"))

    tree = formtree.build(settings)
    return render_template(
        "settings/index.html",
        tree=tree,
        action=url_for("settings.update"),
    )


@bp.route("/new")
def new() -> Any:
    tree = formtree.build(config.build_default_settings())
    return render_template(
        "settings/new.html",
        tree=tree,
        action=url_for("settings.update"),
    )


@bp.route("/update", methods=["POST"])
def update() -> Any:
    ctx = current_ctx()
    tree = formtree.build(config.Settings)
    tree = formtree.validate(tree, request.form)
    if tree.errors > 0:
        return render_template(
            "settings/index.html",
            tree=tree,
            action=url_for("settings.update"),
        )

    data = formtree.unflatten(tree, request.form)
    cfg = load_value(config.Settings, data).normalized()
    ctx.emit(message.SettingsUpdate(settings=cfg))
    return redirect(url_for("settings.index"))



from dataclasses import dataclass
from bussdcc import Message
from ..config import Settings


@dataclass(slots=True, frozen=True)
class SettingsReplaced(Message):
    settings: Settings


@dataclass(slots=True, frozen=True)
class SettingsUpdate(Message):
    settings: Settings


@dataclass(slots=True, frozen=True)
class ConfigChanged(Message):
    pass


@dataclass(slots=True, frozen=True)
class ConfigSaved(Message):
    pass



import json
from pathlib import Path

from bussdcc_framework.codec import load_value, dump_value
from bussdcc_system.model import DeviceSpec

from .config import Config
from .settings import Settings


class ConfigStore:
    def __init__(self, path: str | Path):
        if isinstance(path, str):
            path = Path(path)

        self.path = path
        self.data: Config | None = None

        if path.exists():
            raw = json.loads(path.read_text())

            devices = {
                device_id: DeviceSpec(**spec)
                for device_id, spec in raw.get("devices", {}).items()
            }

            self.data = Config(
                settings=load_value(Settings, raw["settings"]),
                devices=devices,
            )

    def save(self) -> None:
        if self.data is None:
            return

        self.path.parent.mkdir(parents=True, exist_ok=True)

        # dump_value will convert DeviceSpec correctly if codec supports dataclasses
        self.path.write_text(json.dumps(dump_value(self.data), indent=2))



from bussdcc import ContextProtocol, Event, Message, Process

from .. import message


class SettingsProcess(Process):
    name = "settings"

    def handle_event(self, ctx: ContextProtocol, evt: Event[Message]) -> None:
        payload = evt.payload

        if isinstance(payload, message.SettingsReplaced):
            ctx.state.set("settings", payload.settings)
            return

        if isinstance(payload, message.SettingsUpdate):
            ctx.state.set("settings", payload.settings)
            ctx.emit(message.ConfigChanged())



from typing import Any
import secrets

from flask import request, redirect, url_for

from bussdcc import ContextProtocol, Event, Message

from bussdcc_framework.web import FlaskApp, WebInterface as Base

from .blueprints.settings import bp as settings_bp


class WebInterface(Base):
    def register_routes(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        app.secret_key = secrets.token_hex(32)

        @app.route("/")
        def index() -> Any:
            return redirect(url_for("bussdcc_system_devices.index"))

        @app.before_request
        def initial_configuration() -> Any:
            settings = ctx.state.get("settings")
            if settings is not None:
                return

            allowed_endpoints = {
                "settings.new",
                "settings.update",
                "bussdcc_framework_bootstrap.static",
                "bussdcc_framework_formtree.static",
                "bussdcc_framework_client.static",
                "bussdcc_framework_socketio.static",
            }
            if request.endpoint not in allowed_endpoints:
                return redirect(url_for("settings.new"))

        @app.context_processor
        def get_context() -> dict[str, Any]:
            config = ctx.state.get("config")

            return dict(
                settings=ctx.state.get("settings"),
                ri=ctx.state.get("runtime_info", {}),
            )

        app.register_blueprint(settings_bp)

    def handle_event(self, ctx: ContextProtocol, evt: Event[Message]) -> None:
        pass



from dataclasses import dataclass, field, replace
from datetime import datetime
from zoneinfo import ZoneInfo


@dataclass(frozen=True, slots=True)
class ScheduleConfig:
    start_datetime: datetime = field(
        metadata={
            "label": "Start Datetime",
            "group": "Schedule",
            "required": True,
            "help": "Datetime used to calculate age and active schedule step.",
        }
    )


@dataclass(frozen=True, slots=True)
class Settings:
    name: str = field(
        metadata={
            "label": "Name",
            "group": "General",
            "required": True,
            "help": "Unique thermostat instance name.",
        }
    )
    schedule: ScheduleConfig = field(
        metadata={
            "label": "Schedule",
            "group": "Control",
            "required": True,
        }
    )
    timezone: str = field(
        default="America/Chicago",
        metadata={
            "label": "Timezone",
            "group": "General",
            "required": True,
            "help": "IANA timezone name used to normalize input.",
        },
    )

    def normalized(self) -> "Settings":
        tz = ZoneInfo(self.timezone)
        start = self.schedule.start_datetime

        if start.tzinfo is None or start.tzinfo.utcoffset(start) is None:
            start = start.replace(tzinfo=tz)
        else:
            start = start.astimezone(tz)

        return replace(
            self,
            schedule=replace(self.schedule, start_datetime=start),
        )


def default_schedule(start_datetime: datetime) -> ScheduleConfig:
    return ScheduleConfig(
        start_datetime=start_datetime,
    )


def build_default_settings(
    start_datetime: datetime | None = None,
    timezone: str = "America/Chicago",
) -> Settings:
    tz = ZoneInfo(timezone)

    if start_datetime is None:
        start_datetime = datetime.now(tz)
    elif (
        start_datetime.tzinfo is None
        or start_datetime.tzinfo.utcoffset(start_datetime) is None
    ):
        start_datetime = start_datetime.replace(tzinfo=tz)
    else:
        start_datetime = start_datetime.astimezone(tz)

    return Settings(
        name="device-manager-1",
        schedule=default_schedule(start_datetime),
        timezone=timezone,
    ).normalized()



from bussdcc import Service, ContextProtocol, Event, Message
from bussdcc_system import message as system_message

from ..config import ConfigStore, Config
from .. import message


class ConfigService(Service):
    name = "config"

    def __init__(self, data_dir: str) -> None:
        self._data_dir = data_dir

    def _save_config(self, ctx: ContextProtocol) -> None:
        settings = ctx.state.get("settings")
        devices = ctx.state.get("devices", {})

        if settings is None:
            return

        self.cs.data = Config(
            settings=settings,
            devices=devices,
        )
        self.cs.save()

        ctx.emit(message.ConfigSaved())

    def start(self, ctx: ContextProtocol) -> None:
        self.cs = ConfigStore(f"{self._data_dir}/config.json")
        if self.cs.data is None:
            return

        ctx.emit(message.SettingsReplaced(self.cs.data.settings))
        ctx.emit(system_message.DevicesReplaced(self.cs.data.devices))

    def handle_event(self, ctx: ContextProtocol, evt: Event[Message]) -> None:
        payload = evt.payload

        if isinstance(payload, message.ConfigChanged):
            self._save_config(ctx)
            return

        if isinstance(
            payload,
            (
                system_message.DevicesReplaced,
                system_message.DeviceAdded,
                system_message.DeviceConfigUpdate,
                system_message.DeviceDeleted,
            ),
        ):
            self._save_config(ctx)

    def stop(self, ctx: ContextProtocol) -> None:
        self._save_config(ctx)



from .store import ConfigStore
from .config import Config
from .settings import Settings, build_default_settings

__all__ = [
    "ConfigStore",
    "Config",
    "Settings",
    "build_default_settings",
]
